! Some Algorithms for the Conditional Mean and Covariance Matrix
!  John F Monahan, Dept of Statistics, NC State Univ, Sep 2005/Jul 2006
!
! Journal of Statistical Software 16(8), http://www.jstatsoft.org/v16/i08/
!
! Example 1 -- use of sweep operator to compute conditional covariance
!  matrix, mean vector, and determinant
!   
!   computes 
!  a) conditional covariance matrix A1.2 = A11 - A12*inv(A22)*A21
!  b) matrix for conditional mean vector A12*inv(A22)
!  c) determinant of A22
!
      program gsubex1 
      implicit none
      integer, parameter :: n = 3       ! dimension of matrix
      integer, parameter :: nsw = 4     ! number of sweep steps
      real, dimension(n,n) :: A         ! declare matrix A
      integer, dimension(nsw) :: swlist ! holds list of sweep steps
      integer, dimension(n) :: cstat,clist,ulist 
      integer i,j,k,kk,lclist,lulist,idet
      real det
!                   interface block
      interface
      subroutine sweepd(a,m,k,det,idet)
      real, dimension(:,:), intent(in out) :: a
      integer, intent(in) :: m,k
      integer, intent(in out) :: idet
      real, intent(in out) :: det
      end subroutine sweepd
      end interface
!
  22  format('Sweep tableau')
  23  format(5f15.8)
  24  format(' after step ',i2,'  sweeping k=',i2,' determinant=', &
     &  f8.4,'*(2**(',i3,') )') 
  25  format('inv(A22)*A21',5x,5i8)
  26  format(14x,i3,5f8.4)
  27  format('   A1.2',10x,5i8)
  28  format(' Distribution of components ',3i3)
  29  format(10x,'conditional on components ',3i3)
  30  format('Stopped -- computationally singular matrix')
!                   matrix A
      data A/3., 1., 0., 1., 4., 2., 0., 2., 6./
      data cstat/ 0, 0, 0/              ! condition on none to start
      data ulist/ 1, 2, 3/, lulist/ 3/
      data clist/ 0, 0, 0/, lclist/ 0/
      data swlist/ 2, 3, 2, 1/
!
      open( unit=6, file='gsubex1.out' )
!
!                   initialize determinant 
      det = 1.
      idet = 0
!                   write out matrix A
      write(6,22)
      do  i = 1,n
      write(6,23) (A(i,j),j=1,n)
      end do  ! loop on i
      write(6,24) 0,0,det,idet
!                   now print A1.2
      write(6,27) (ulist(j),j=1,lulist)
      do i = 1, lulist
      write(6,26) ulist(i),(A(ulist(i),ulist(j)),j=1,lulist)
      end do  ! loop on i
!                   sweep variable k
      do kk = 1,nsw
!
!                   now sweep variable k
      k = swlist(kk) 
      cstat(k) = 1 - cstat(k)           ! condition/uncondition Z(k)
      call sweepd(A,n,k,det,idet)
!                   check if computationally singular
      if( idet .lt. -2000000000 ) then
                                  write(6,30)
                                  stop
                                  end if ! ( idet .lt. -2000000000 )
      write(6,22)
      do i = 1,n
      write(6,23) (A(i,j),j=1,n)
      end do  ! loop on i
      write(6,24) kk,k,det,idet 
!                   print A12*inv(A22) and A1.2
!                   find list of conditioned/not components first
      lclist = 0
      lulist = 0
      do i = 1, n
      if( cstat(i) .eq. 0 ) then
                              lulist = lulist + 1
                              ulist(lulist) = i
                            else
                              lclist = lclist + 1
                              clist(lclist) = i 
                            end if ! ( cstat(i) .eq. 0 )
      end do  ! loop on i
!                   conditioning on lclist components
      write(6,28) (ulist(i),i=1,lulist)
      write(6,29) (clist(i),i=1,lclist)
      if( lulist .eq. 0 ) cycle         ! nothing to print
!                   write out A12*inv(A22) and A1.2
!                   first A12*inv(A22)
      write(6,25) (clist(j),j=1,lclist)
      do i = 1, lulist
      write(6,26) ulist(i),(-A(ulist(i),clist(j)),j=1,lclist)
      end do  ! loop on i
!                   now print A1.2
      write(6,27) (ulist(j),j=1,lulist)
      do i = 1, lulist
      write(6,26) ulist(i),(A(ulist(i),ulist(j)),j=1,lulist)
      end do  ! loop on i
!                   done with printing
      end do  ! loop on kk
      stop
      end program gsubex1
      subroutine sweepd(a,m,k,det,idet)
! sweep operator on the sum of squares and cross products mx A
!  a  real matrix m by m  (dummy dimensioned -- use INTERFACE block)
!  m  integer size of matrix p+1 for one dependent and p regressors
!  k  integer variable to be swept
!
! update of SWEEP to include determinant calculation
!  use of ADJUST presupposes positive definite matrix
      implicit none
      real, dimension(:,:), intent(in out) :: a
      integer, intent(in) :: m,k
      integer, intent(in out) :: idet
      real, intent(in out) :: det
      real b,d
      integer i,j
!                   sweeping in/out variable k
      d = a(k,k)
      det = det * d
      call adjust(det,idet)
      do j = 1,m
      a(k,j) = a(k,j)/d
      end do  ! loop on j
!                   fixed row k, now rest of matrix
      do i = 1,m
      if( i .ne. k ) then
      b = a(i,k) 
      do j = 1,m
      a(i,j) = a(i,j) - b*a(k,j)
      end do  ! loop on j
      a(i,k) = -b/d
                     END if
      end do  ! loop on i
      a(k,k) = 1.d0/d 
!                   done
      return
      end subroutine sweepd
      SUBROUTINE ADJUST(D,I)
!  NORMALIZES D WHILE KEEPING CONSTANT VALUE OF D * (2**I)
      INTEGER, PARAMETER :: IBIG = -2147483644
      REAL, INTENT (IN OUT) :: D
      INTEGER, INTENT(IN OUT) :: I
      IF ( D .GT. 0.0 ) THEN
        DO WHILE ( D .LT. 1.0 )
        D = D*16.
        I = I-4
        ENDDO
        DO WHILE ( D .GT. 16.0 )
        D = D/16.
        I = I+4
        ENDDO
                        ELSE
      I = IBIG      ! IF D < 0 THEN I = - BIG
                        ENDIF
      RETURN
      END SUBROUTINE ADJUST

