! Some Algorithms for the Conditional Mean and Covariance Matrix
!  John F Monahan, Dept of Statistics, NC State Univ, Sep 2005/Jul 2006
!
! Journal of Statistical Software 16(8), http://www.jstatsoft.org/v16/i08/
!
! Example 2 -- reverse Cholesky factorization with Givens transforms to
!  compute conditional covariance matrix, mean vector, and determinant
!
!   computes 
!  a) conditional covariance matrix A1.2 = A11 - A12*inv(A22)*A21
!  b) matrix for conditional mean vector A12*inv(A22)
!  c) determinant of A22
!
      program gsubex2
      implicit none
      integer, parameter :: n = 4       ! dimension of matrix
      real, dimension(n,n) :: A         ! declare matrix A
      real, dimension( (n*(n+1))/2 ) :: At,Atp  ! triangular form
      real, dimension(n) :: v           ! temp storage 
      integer, dimension(n) :: p        ! permutation vector
      integer i,j,k,kmin,l,m,idet,glist
      real det,rnu,sa,sb
      integer, parameter :: glistmax = (n*(n+1))/2   ! max # of G's
      integer, parameter :: s = 2       ! length of subset 
      integer, dimension(glistmax,2) :: pg   ! k's and l's
      real, dimension(glistmax,2) :: G       ! gamma's and sigma's
!                   interface block
      interface
      subroutine sweepd(a,m,k,det,idet)
      real, dimension(:,:), intent(in out) :: a
      integer, intent(in) :: m,k
      integer, intent(in out) :: idet
      real, intent(in out) :: det
      end subroutine sweepd
      end interface
!
  21  format('   A1.2',10x,5i8)
  22  format(14x,i3,5f8.4)
  23  format(5f15.8)
  24  format(' after sweeping row/cols ',i2,' and ',i2,' det=', &
     &  f12.4,'*(2**(',i3,') )') 
  25  format('inv(A22)*A21',5x,5i8)
  26  format(2X,'determinant is ',F9.6,' * ( 2**',I6,' )')
  27  format('  factored matrix -- upper triangular U ')
  28  format(' permutation ',15i3)
  30  format('Stopped -- computationally singular matrix')
  31  format(' original matrix A ')
  32  format(" permuted matrix PAP'")
  33  format(" new upper triangular U = P*U*G1*G2 = factor of (PAP')")
!                   matrix A
      data A/25., 11., 0., 4.,  11., 19., 7., 2.,  0., 7., 5., 2.,    &
     &        4., 2., 2., 4./
      data At/25., 11., 19., 0., 7., 5., 4., 2., 2., 4./
      data Atp/25., 4., 4., 11., 2., 19., 0., 2., 7., 5./
      data p/ 1, 4, 2, 3 /
!
      open( unit=6, file='gsubex2.out' )
!
!                   initialize determinant 
      det = 1.
      idet = 0
!                   write out matrix A
      write(6,31)
      do  i = 1,n
      write(6,23) (A(i,j),j=1,n)
      end do  ! loop on i
!                   two sweeps, second and third
      call sweepd(A,n,2,det,idet)
      call sweepd(A,n,3,det,idet)
      write(6,24) 2,3,det,idet 
      do i = 1,n
      write(6,23) (A(i,j),j=1,n)
      end do  ! loop on i
!                   write out A1.2
      write(6,21) 1, 4
      write(6,22) 1, A(1,1), A(1,4)
      write(6,22) 4, A(4,1), A(4,4)
!                   write out A12*inv(A22)
      write(6,25) 2, 3
      write(6,22) 1, -A(1,2), -A(1,3)
      write(6,22) 4, -A(4,2), -A(4,3)
!                   factor permuted matrix Atp (solution)
      write(6,32)
      do j = 1, n
      write(6,23) (Atp( (j*(j-1))/2 + i ), i = 1, (j-1) ),  &
     &                   (Atp( (i*(i-1))/2 + j ), i = j, n )
      end do  ! loop on j
      call rchlzky(Atp,n,det,idet)
      write(6,26) det,idet
!                   check if computationally singular
      if( idet .lt. -2000000000 ) then
                                  write(6,30)
                                  stop
                                  end if ! ( idet .lt. -2000000000 )
      write(6,27)
      do j = 1, n
      write(6,23) (0.,i=1,(j-1)),(Atp( (i*(i-1))/2 + j ), i = j, n )
      end do  ! loop on j
!                   reuse later -- wipe out for debugging
      Atp = -2.
!                   factor original matrix A
      write(6,31)
      call rchlzky(At,n,det,idet)
      write(6,26) det,idet
!                   check if computationally singular
      if( idet .lt. -2000000000 ) then
                                  write(6,30)
                                  stop
                                  end if ! ( idet .lt. -2000000000 )
      write(6,27)
      do j = 1, n
      write(6,23) (0.,i=1,(j-1)),(At( (i*(i-1))/2 + j ), i = j, n )
      end do  ! loop on j
      write(6,28) p
!                   now permute and work back
      glist = 0
      do i = n, 1, -1
!                   copy column into v
      v = 0.
      do j = p(i), n
      v(j) = At( (j*(j-1))/2 + p(i) )   ! copy col p(i) into v
      end do  ! loop on j
      kmin = p(i)
!                   apply old G's to v
      if ( glist .gt. 0 ) then
      do m = 1, glist
      k = pg(m,1)
      kmin = min( kmin, k )   ! smallest k -- first nonzero in v
      l = pg(m,2)             ! k < l
      sa = G(m,1)*v(k) - G(m,2)*v(l)
      sb = G(m,1)*v(l) + G(m,2)*v(k)
      v(k) = sa
      v(l) = sb
      end do  ! loop on m -- old G's
                          endif ! ( glist .gt. 0 )
!                   create new G's in putting zeros in v(kmin),...,v(p(i))
      if( kmin .lt. i ) then  ! is there any need to add zeros?
         do k = kmin, i-1
         glist = glist + 1
         pg(glist,1) = k
         pg(glist,2) = i
         call rot734( v(i), v(k), G(glist,1), G(glist,2), rnu )
!                             apply to v
         v(k) = 0.
         v(i) = rnu
         end do  ! loop on k
                        endif ! ( kmin .lt. i )
!                             copy v into col i
      do j = i, n
      Atp( (j*(j-1))/2 + i ) = v(j)       ! reuse Atp
      end do  ! loop on j
!                             done with column i
      end do  ! loop on i
!                             write out results
      write(6,32)
      write(6,33)
      do j = 1, n
      write(6,23) (0.,i=1,(j-1)),(Atp( (i*(i-1))/2 + j ), i = j, n )
      end do  ! loop on j
!
!  take upper triangular U and construct pieces for conditional 
!    distribution, A12*inv(A22) and A1.2
!  for example 1,4 given 2,3 but in permuted form, 1,2 given 3,4
!   in general 1, ..., s given s+1, ..., n
!    
!                     first A1.2 = (U1)*(U1)'
      write(6,21) (i, i = 1, s) 
      do i = 1, s
      do j = 1, s
      sa = 0.
      do k = max(i,j), s
      sa = sa + Atp( (k*(k-1))/2 + i )*Atp( (k*(k-1))/2 + j )
      end do  ! loop on k
      A(i,j) = sa
      end do  ! loop on j
!                   write out A1.2
      write(6,22) i, (A(i,j), j=1, s) 
      end do  ! loop on i 
!
!                     A12*inv(A22) = U2*inv(U3)
      write(6,25) (j, j = s+1, n)
      do i = 1, s
      do j = s+1, n
      A(i,j) = Atp( (j*(j-1))/2 + i )             ! U2
      end do  ! loop on j
!                     solve lower triangular system (U3')() =  U2(i,.)'
      do j = s+1, n
      if ( j .gt. s+1 ) then
        do k = s+1, j-1
        A(i,j) = A(i,j) - Atp( (j*(j-1))/2 + k )*A(i,k)
        end do  ! loop on k
                        end if ! ( j .gt. s+1 )
      A(i,j) = A(i,j) / Atp( (j*(j+1))/2 )
      end do  ! loop on j 
      write(6,22) i, (A(i,j), j=s+1, n)
      end do  ! loop on i
      stop
      end program gsubex2
      subroutine sweepd(a,m,k,det,idet)
! sweep operator on the sum of squares and cross products mx A
!  a  real matrix m by m  (dummy dimensioned -- use INTERFACE block)
!  m  integer size of matrix p+1 for one dependent and p regressors
!  k  integer variable to be swept
!
! update of SWEEP to include determinant calculation
!  use of ADJUST presupposes positive definite matrix
      implicit none
      real, dimension(:,:), intent(in out) :: a
      integer, intent(in) :: m,k
      integer, intent(in out) :: idet
      real, intent(in out) :: det
      real b,d
      integer i,j
!                   sweeping in/out variable k
      d = a(k,k)
      det = det * d
      call adjust(det,idet)
      do j = 1,m
      a(k,j) = a(k,j)/d
      end do  ! loop on j
!                   fixed row k, now rest of matrix
      do i = 1,m
      if( i .ne. k ) then
      b = a(i,k) 
      do j = 1,m
      a(i,j) = a(i,j) - b*a(k,j)
      end do  ! loop on j
      a(i,k) = -b/d
                     END if
      end do  ! loop on i
      a(k,k) = 1.d0/d 
!                   done
      return
      end subroutine sweepd
      SUBROUTINE ADJUST(D,I)
!  NORMALIZES D WHILE KEEPING CONSTANT VALUE OF D * (2**I)
      INTEGER, PARAMETER :: IBIG = -2147483644
      REAL, INTENT (IN OUT) :: D
      INTEGER, INTENT(IN OUT) :: I
      IF ( D .GT. 0.0 ) THEN
        DO WHILE ( D .LT. 1.0 )
        D = D*16.
        I = I-4
        ENDDO
        DO WHILE ( D .GT. 16.0 )
        D = D/16.
        I = I+4
        ENDDO
                        ELSE
      I = IBIG      ! IF D < 0 THEN I = - BIG
                        ENDIF
      RETURN
      END SUBROUTINE ADJUST
      SUBROUTINE RCHLZKY(A,N,DET,IDET)
!  RCHLSKY COMPUTES THE reverse CHOLESKY (SQUARE-ROOT)  FACTORIZATION
!   A = ( L - TRANSPOSE )*L   WHERE L IS LOWER TRIANGULAR
! ***  USES SYMMETRIC STORAGE FORMAT  ***
!
!     A IS OVERWRITTEN BY L IN ITS LOWER TRIANGULAR PART
!  SUBROUTINE ADJUST KEEPS DETERMINANT FROM EXPLODING USING
!  1 LE DET LE 16   AND   DETERMINANT OF A IS DET*2**IDET
!
!  J F MONAHAN (DEC,1983) DEPT OF STAT, N C S U, RALEIGH, N C 27650 USA
!      REVISED JULY 1999, APRIL 2000 FOR FORTRAN 95
!      reverse Cholesky form -- August 2005
      IMPLICIT NONE
      INTEGER, INTENT(IN) :: N
      REAL, DIMENSION( (N*(N+1))/2 ), INTENT(IN OUT) :: A
      REAL, INTENT(OUT) :: DET
      INTEGER, INTENT(OUT) :: IDET
      REAL T,S
      INTEGER I,J,K,ip1,jp1
!                                    INITIALIZE FOR DETERMINANT
      DET = 1.                                                        
      IDET = 0                                                          
!                                      DO I-TH ROW
      DO i = n, 1, -1                  ! reverse order
      T = A( (I*(I+1))/2 )
!                                      FIRST ROW IS TRIVIAL
      IF(i .lt. n) THEN
        ip1 = i+1
        DO J = n, ip1, -1              ! reverse order
!                                      WORK ON (j,i)-TH ELEMENT
        S = A( (j*(j-1))/2 + i )
        IF(j .lt. n) THEN
          jp1 = j + 1 
          DO K = n, jp1, -1            ! reverse order 
          S = S - A( (k*(k-1))/2 + j ) * A( (k*(k-1))/2 + i )
          ENDDO     ! LOOP ON K
                     ENDIF
        A( (j*(j-1))/2 + i )= S / A( (J*(J+1))/2 )
!                                 WORK ON DIAGONAL ELEMENT
        T = T - A( (j*(j-1))/2 + i ) * A( (j*(j-1))/2 + i )
        ENDDO       ! LOOP ON J
!                                 FINISHED WITH LOOP ON J
!                                 UPDATE DETERMINANT WITH DIAGONAL
                   ENDIF
      DET = DET * T
!                                 KEEP DET FROM EXPLODING
      CALL ADJUST(DET,IDET)                                             
!                                 ABORT IF DET IS NEGATIVE OR ZERO
      IF(IDET.LT.-2000000000) RETURN                                    
!                                 FINISH A POSITIVE DIAGONAL
      A( (I*(I+1))/2 ) = SQRT(T)
      ENDDO         ! LOOP ON I
      RETURN                                                            
      END SUBROUTINE RCHLZKY
      SUBROUTINE ROT734(ALPHA,BETA,GAMA,SIGMA,RNU)
!  GIVEN ALPHA AND BETA, PRODUCES GIVENS TRANSFORMATION VALUES
!
!      (  GAMA       SIGMA  )  (  ALPHA  )       (  RNU  )
!      (                    )  (         )   =   (       )
!      ( -SIGMA       GAMA  )  (  BETA   )       ( ZERO  )
!
!  ALGORITHM 7.3.4 FROM STEWART
!
      IMPLICIT NONE
      REAL, INTENT(IN) :: ALPHA,BETA
      REAL, INTENT(OUT) :: GAMA,SIGMA,RNU
      REAL ETA,DELTA
!
      ETA = MAX( ABS(ALPHA), ABS(BETA) )
!                     IF BOTH ALPHA AND BETA ARE ZERO THEN
!                     USE IDENTITY TRANSFORMATION
      IF( ETA .EQ. 0. ) then
!                     SET UP IDENTITY TRANSFORMATION IF NO WORK DONE
      RNU = 0.
      GAMA = 1.                                                         
      SIGMA = 0.                                                        
      RETURN
                        else
      DELTA = (ALPHA/ETA)*(ALPHA/ETA) + (BETA/ETA)*(BETA/ETA)           
      DELTA = SQRT(DELTA)                                               
      RNU = ETA * DELTA                                                 
      GAMA = ALPHA / RNU                                                
      SIGMA = BETA / RNU
                        END if
      RETURN                                                            
      END SUBROUTINE ROT734


